import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_sequence import  \
    DATA_simple_seq, DATA_bool_seq, SCHEMA_two_elem_seq_with_extension_marker_present_or_not, \
    SCHEMA_seq_with_only_extension_marker_present_or_not, SCHEMA_seq_with_elements_in_extension_addition, \
    SCHEMA_seq_with_some_optional_elements_in_root, SCHEMA_seq_with_only_optional_elements_in_root, \
    SCHEMA_seq_with_default_elements_in_root, SCHEMA_sequence_of_boolean, DATA_octet_seq, SCHEMA_sequence_of_octetstring, \
    SCHEMA_sequence_of_bitstring, DATA_bitstring_seq, DATA_enumerated_seq, SCHEMA_sequence_of_enumerated, \
    SCHEMA_sequence_with_extension_addition_groups_1, SCHEMA_sequence_with_extension_addition_groups_2, \
    SCHEMA_sequence_with_extension_addition_groups_3, DATA_extension_addition_groups_seq


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_two_elem_seq_with_extension_marker_present_or_not(extensionMarker_value=False), '0300870702FCB3',
        DATA_simple_seq(SCHEMA_two_elem_seq_with_extension_marker_present_or_not, i0_is=True,
                        i1_is=True, i2_is=False, i3_is=False, i4_is=False, ext_mark=False)),
    (SCHEMA_two_elem_seq_with_extension_marker_present_or_not(extensionMarker_value=True), '000300870702FCB3',
        DATA_simple_seq(SCHEMA_two_elem_seq_with_extension_marker_present_or_not, i0_is=True,
                        i1_is=True, i2_is=False, i3_is=False, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_only_extension_marker_present_or_not(extensionMarker_value=False), '0300870702FCB300178003030D4028',
        DATA_simple_seq(SCHEMA_seq_with_only_extension_marker_present_or_not, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=True, ext_mark=False)),
    (SCHEMA_seq_with_only_extension_marker_present_or_not(extensionMarker_value=True), '000300870702FCB300178003030D4028',
        DATA_simple_seq(SCHEMA_seq_with_only_extension_marker_present_or_not, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=True, ext_mark=True)),

])
def test_no_extension_no_default_no_optional_components_present_sequence_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707E00302FCB3020017058003030D400128',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=True, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707C00302FCB3020017058003030D40',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707800302FCB3020017',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=False, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707000302FCB3',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=False, i3_is=False, i4_is=False, ext_mark=True)),
])
def test_extension_components_are_present_but_no_default_no_optional_sequence_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_seq_with_some_optional_elements_in_root(extensionMarker_value=False), '0003008707001728',
        DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True,
                        i1_is=False, i2_is=True, i3_is=False, i4_is=True)),
    (SCHEMA_seq_with_some_optional_elements_in_root(extensionMarker_value=False), '800300870702FCB3001728',
        DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=False, i4_is=True)),
    (SCHEMA_seq_with_some_optional_elements_in_root(extensionMarker_value=False), '400300870700178003030D4028',
        DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True,
                        i1_is=False, i2_is=True, i3_is=True, i4_is=True)),
    (SCHEMA_seq_with_some_optional_elements_in_root(extensionMarker_value=False), 'C00300870702FCB300178003030D4028',
        DATA_simple_seq(SCHEMA_seq_with_some_optional_elements_in_root, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
    (SCHEMA_seq_with_only_optional_elements_in_root(extensionMarker_value=False), 'C00300870702FCB3',
        DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=True,
                        i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_seq_with_only_optional_elements_in_root(extensionMarker_value=False), '8003008707',
        DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=True,
                        i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_seq_with_only_optional_elements_in_root(extensionMarker_value=False), '4002FCB3',
        DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=False,
                        i1_is=True, i2_is=False, i3_is=False, i4_is=False)),
    (SCHEMA_seq_with_only_optional_elements_in_root(extensionMarker_value=False), '00',
        DATA_simple_seq(SCHEMA_seq_with_only_optional_elements_in_root, i0_is=False,
                        i1_is=False, i2_is=False, i3_is=False, i4_is=False)),
])
def test_optional_components_present_in_extension_root_sequence_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_seq_with_default_elements_in_root(extensionMarker_value=False), '0003008707001728',
        DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True,
                        i1_is=False, i2_is=True, i3_is=False, i4_is=True)),
    (SCHEMA_seq_with_default_elements_in_root(extensionMarker_value=False), '800300870702FCB3001728',
        DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=False, i4_is=True)),
    (SCHEMA_seq_with_default_elements_in_root(extensionMarker_value=False), '400300870700178003030D4028',
        DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True,
                        i1_is=False, i2_is=True, i3_is=True, i4_is=True)),
    (SCHEMA_seq_with_default_elements_in_root(extensionMarker_value=False), 'C00300870702FCB300178003030D4028',
        DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=True)),
])
def test_default_components_with_none_default_values_present_in_extension_root_sequence_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_seq_with_default_elements_in_root(extensionMarker_value=False), '0003008707001728',    # no other test is needed here
        DATA_simple_seq(SCHEMA_seq_with_default_elements_in_root, i0_is=True,                      # as default values are never encoded
                        i1_is=False, i2_is=True, i3_is=False, i4_is=True)),                        # so they never will be decoded
])
def test_default_components_with_default_values_present_in_filled_schema_are_not_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '0003008707',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=False, i2_is=False, i3_is=False, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707000302FCB3',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=False, i3_is=False, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707800302FCB3020017',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=False, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707C00302FCB3020017058003030D40',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=False, ext_mark=True)),
    (SCHEMA_seq_with_elements_in_extension_addition(extensionMarker_value=True), '800300870707E00302FCB3020017058003030D400128',
        DATA_simple_seq(SCHEMA_seq_with_elements_in_extension_addition, i0_is=True,
                        i1_is=True, i2_is=True, i3_is=True, i4_is=True, ext_mark=True)),
])
def test_elements_in_extension_addition_may_not_be_present_even_if_no_optional_in_schema_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_sequence_of_boolean(), '0000', DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=False)),
    (SCHEMA_sequence_of_boolean(), '0001', DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=False)),    # Other version
    (SCHEMA_sequence_of_boolean(), '001F', DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=False)),    # Other version
    (SCHEMA_sequence_of_boolean(), '7FE0', DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=False)),
    (SCHEMA_sequence_of_boolean(), '7FFF', DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=False)),     # Other version
    (SCHEMA_sequence_of_boolean(), '8002BFF801000100010001000100010001000100010001000100',
        DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=True, addition_val=False)),
    (SCHEMA_sequence_of_boolean(), '8002BFF801800180018001800180018001800180018001800180',
        DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=False, addition_present=True, addition_val=True)),
    (SCHEMA_sequence_of_boolean(), 'FFE2BFF801000100010001000100010001000100010001000100',
        DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=True, addition_val=False)),
    (SCHEMA_sequence_of_boolean(), 'FFE2BFF801800180018001800180018001800180018001800180',
        DATA_bool_seq(SCHEMA_sequence_of_boolean, root_val=True, addition_present=True, addition_val=True)),
])
def test_sequence_of_boolean_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_sequence_of_octetstring(), 'E002DEAD02BEEF00ABCD01AA02BBB002CCCC03DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=True, o5_is=True)),
    (SCHEMA_sequence_of_octetstring(), '6002DEAD00ABCD01AA02BBB002CCCC03DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=True, o5_is=True)),
    (SCHEMA_sequence_of_octetstring(), 'A002DEAD02BEEF00ABCD01AA02CCCC03DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=False, o5_is=True)),
    (SCHEMA_sequence_of_octetstring(), 'C002DEAD02BEEF00ABCD01AA02BBB003DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=True, o5_is=False)),
    (SCHEMA_sequence_of_octetstring(), '2002DEAD00ABCD01AA02CCCC03DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=False, o5_is=True)),
    (SCHEMA_sequence_of_octetstring(), '4002DEAD00ABCD01AA02BBB003DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=True, o5_is=False)),
    (SCHEMA_sequence_of_octetstring(), '8002DEAD02BEEF00ABCD01AA03DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=True, o4_is=False, o5_is=False)),
    (SCHEMA_sequence_of_octetstring(), '0002DEAD00ABCD01AA03DDDDD0',
        DATA_octet_seq(SCHEMA_sequence_of_octetstring, o1_is=False, o4_is=False, o5_is=False)),
])
def test_sequence_of_octetstring_no_extension_marker_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_sequence_of_bitstring(), 'F00CF0F010DEAD10BEEF10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=True, b9_is=True)),
    (SCHEMA_sequence_of_bitstring(), 'B00CF0F010DEAD10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=True, b9_is=True)),
    (SCHEMA_sequence_of_bitstring(), 'D00CF0F010DEAD10BEEFF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=False, b9_is=True)),
    (SCHEMA_sequence_of_bitstring(), 'F00CF0F010DEAD10BEEF10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=True, b9_is=False)),
    (SCHEMA_sequence_of_bitstring(), '900CF0F010DEADF810FEED0014F0F0F0F0F010BBBB14CCCCC05C0280C003F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=False, b9_is=True)),
    (SCHEMA_sequence_of_bitstring(), 'B00CF0F010DEAD10ABCDF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=True, b9_is=False)),
    (SCHEMA_sequence_of_bitstring(), 'D00CF0F010DEAD10BEEFF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=True, b3_is=False, b9_is=False)),
    (SCHEMA_sequence_of_bitstring(), '900CF0F010DEADF810FEED0014F0F0F0F0F010BBBB14CCCCC04C03F0F0F00418DDDDDD',
        DATA_bitstring_seq(SCHEMA_sequence_of_bitstring, b2_is=False, b3_is=False, b9_is=False)),
])
def test_sequence_of_bitstring_with_extension_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_sequence_of_enumerated(), 'EC0A80E001800100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=True, e4_is=True)),
    (SCHEMA_sequence_of_enumerated(), 'AA80E001800100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=True, e4_is=True)),
    (SCHEMA_sequence_of_enumerated(), 'CC0A0701800100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=False, e4_is=True)),
    (SCHEMA_sequence_of_enumerated(), 'EC0A80A00100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=True, e4_is=False)),
    (SCHEMA_sequence_of_enumerated(), '8A0701800100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=False, e4_is=True)),
    (SCHEMA_sequence_of_enumerated(), 'AA80A00100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=False, e3_is=True, e4_is=False)),
    (SCHEMA_sequence_of_enumerated(), 'CC0A050100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated, e1_is=True, e3_is=False, e4_is=False)),
    (SCHEMA_sequence_of_enumerated(), '8A050100', DATA_enumerated_seq(SCHEMA_sequence_of_enumerated,  e1_is=False, e3_is=False, e4_is=False)),
])
def test_sequence_of_enumerated_with_extension_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_sequence_with_extension_addition_groups_1(), '00019901AA', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_1, s0='99', s1='AA')),
    (SCHEMA_sequence_with_extension_addition_groups_1(), '80019901AA07000401BB01CC', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_1, s0='99', s1='AA', s2='BB', s3='CC')),
    (SCHEMA_sequence_with_extension_addition_groups_1(), '80019901AA07800401BB01CC0201DD', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_1, s0='99', s1='AA', s2='BB', s3='CC', s4='DD')),
    (SCHEMA_sequence_with_extension_addition_groups_1(), '80019901AA07C00401BB01CC0201DD0201EE', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_1, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE')),
    (SCHEMA_sequence_with_extension_addition_groups_1(), '80019901AA07E00401BB01CC0201DD0201EE0201FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_1, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '000199', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905000201AA', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905800201AA038001BB', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s2='BB')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905800201AA034001CC', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s3='CC')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905800201AA05C001BB01CC', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s2='BB', s3='CC')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905C00201AA05C001BB01CC078001DD01EE01FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905C00201AA05C001BB01CC050001DD01FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905400201AA078001DD01EE01FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s4='DD', s5='EE', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_2(), '80019905400201AA050001DD01FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_2, s0='99', s1='AA', s4='DD', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '000199', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019907000201AA', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s1='AA')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019907800201AA0201BB', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s1='AA', s2='BB')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019906800201BB', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s2='BB')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019907C00201AA0201BB050001CC01DD', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s1='AA', s2='BB', s3='CC', s4='DD')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019907C00201AA0201BB078001CC01DD01EE', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019907E00201AA0201BB078001CC01DD01EE038001FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s5='EE', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019907E00201AA0201BB050001CC01DD038001FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s1='AA', s2='BB', s3='CC', s4='DD', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019906E00201BB050001CC01DD038001FF', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s2='BB', s3='CC', s4='DD', s6='FF')),
    (SCHEMA_sequence_with_extension_addition_groups_3(), '80019906C00201BB050001CC01DD', DATA_extension_addition_groups_seq(
        SCHEMA_sequence_with_extension_addition_groups_3, s0='99', s2='BB', s3='CC', s4='DD')),
])
def test_sequence_with_extension_addition_groups_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema) == value


import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.test.per.encoder.test_per_encode_choice import SCHEMA_single_item_choice, SCHEMA_no_extension_simple_choice, \
    SCHEMA_reverse_no_extension_simple_choice, SCHEMA_reverse_with_extension_simple_choice, \
    SCHEMA_selected_choice_from_extension_addition_more_elems, SCHEMA_selected_choice_from_extension_addition_one_elem, \
    SCHEMA_single_item_choice_with_extension, SCHEMA_with_extension_simple_choice, SCHEMA_selected_choice_from_extension_addition_groups_1, \
    SCHEMA_selected_choice_from_extension_addition_groups_2, DATA_choice


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_single_item_choice(), '010A', DATA_choice(SCHEMA_single_item_choice(), identifier='i0', value=10)),
    (SCHEMA_single_item_choice(), '030186A0', DATA_choice(SCHEMA_single_item_choice(), identifier='i0', value=100000)),
    (SCHEMA_no_extension_simple_choice(), '00010A', DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i0', value=10)),
    (SCHEMA_no_extension_simple_choice(), '00030186A0', DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i0', value=100000)),
    (SCHEMA_no_extension_simple_choice(), '80010A', DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i1', value=10)),
    (SCHEMA_no_extension_simple_choice(), '80030186A0', DATA_choice(SCHEMA_no_extension_simple_choice(), identifier='i1', value=100000)),
    (SCHEMA_reverse_no_extension_simple_choice(), '80010A', DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i0', value=10)),
    (SCHEMA_reverse_no_extension_simple_choice(), '80030186A0', DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i0', value=100000)),
    (SCHEMA_reverse_no_extension_simple_choice(), '00010A', DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i1', value=10)),
    (SCHEMA_reverse_no_extension_simple_choice(), '00030186A0', DATA_choice(SCHEMA_reverse_no_extension_simple_choice(), identifier='i1', value=100000)),
])
def test_no_extension_marker_choice_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_single_item_choice_with_extension(), '000108', DATA_choice(SCHEMA_single_item_choice_with_extension(), identifier='i0', value=8)),
    (SCHEMA_single_item_choice_with_extension(), '0003010B90', DATA_choice(SCHEMA_single_item_choice_with_extension(), identifier='i0', value=68496)),
    (SCHEMA_with_extension_simple_choice(), '0002FF7E', DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i0', value=-130)),
    (SCHEMA_with_extension_simple_choice(), '00025B9F', DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i0', value=23455)),
    (SCHEMA_with_extension_simple_choice(), '4001FF', DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i1', value=-1)),
    (SCHEMA_with_extension_simple_choice(), '4003FE7960', DATA_choice(SCHEMA_with_extension_simple_choice(), identifier='i0', value=-100000)),
    (SCHEMA_reverse_with_extension_simple_choice(), '40010A', DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i0', value=10)),
    (SCHEMA_reverse_with_extension_simple_choice(), '40030186A0', DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i0', value=100000)),
    (SCHEMA_reverse_with_extension_simple_choice(), '00010A', DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i1', value=10)),
    (SCHEMA_reverse_with_extension_simple_choice(), '00030186A0', DATA_choice(SCHEMA_reverse_with_extension_simple_choice(), identifier='i1', value=100000)),
])
def test_only_extension_marker_preset_choice_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '00010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i0', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '04010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i1', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '0C010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i3', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '18010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i6', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '24010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i9', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '30010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i12', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '3C010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i15', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '40010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i16', value=10)),
])
def test_extension_present_but_choice_from_root_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_selected_choice_from_extension_addition_one_elem(), '8002010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i2', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_one_elem(), '800302F830',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i2', value=-2000)),
    (SCHEMA_selected_choice_from_extension_addition_one_elem(), '8102010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i3', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_one_elem(), '810302F830',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_one_elem(), identifier='i3', value=-2000)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '80020100',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i17', value=0)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '8002010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i17', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '810201F6',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i18', value=-10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '8202010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i19', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '8502010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i22', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '8702010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i24', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '8902010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i26', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '8C02010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i29', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '9002010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i33', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '9102010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i34', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_more_elems(), '9202010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_more_elems(), identifier='i35', value=10)),
])
def test_choice_from_extension_addition_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '00010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i0', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '40010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i1', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8002010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i2', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8102010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i3', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8202010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i4', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8302010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i5', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8402010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i6', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8502010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i7', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8602010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i8', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_1(), '8702010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_1(), identifier='i9', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_2(), '00010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i0', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_2(), '40010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i1', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_2(), '8002010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i5', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_2(), '8102010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i7', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_2(), '8202010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i8', value=10)),
    (SCHEMA_selected_choice_from_extension_addition_groups_2(), '8302010A',
        DATA_choice(SCHEMA_selected_choice_from_extension_addition_groups_2(), identifier='i9', value=10)),
])
def test_choice_from_extension_addition_groups_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class MyEnum(EnumeratedType):
    subtypeSpec = ExtensionMarker(True)
    enumerationRoot = NamedValues(
        ('zero', 0),
        ('one', 1),
        ('two', 2),
        ('three', 3),
    )
    extensionAddition = NamedValues(
        ('four', 4),
        ('five', 5),
    )
    namedValues = enumerationRoot + extensionAddition


class ThirdSequence(SequenceOfType):
    componentType = OctetStringType()


class MyChoice(ChoiceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('c0', IntegerType()),
        NamedType('c1', ThirdSequence()),
        NamedType('c2', BooleanType()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('c3', IntegerType()),
        NamedType('c4', BooleanType()),
    )
    componentType = rootComponent + extensionAddition


class YetOtherSeq(SequenceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('myEnum', MyEnum()),
        NamedType('myChoice', MyChoice()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('add3', OctetStringType()),
        NamedType('add4', IntegerType()),
    )
    componentType = rootComponent + extensionAddition


class OtherSequence(SequenceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('yetOtherSeq', YetOtherSeq()),
        NamedType('myBool', BooleanType()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('add2', BitStringType()),
    )
    componentType = rootComponent + extensionAddition


class MySeq(SequenceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('int0', IntegerType()),
        NamedType('otherSeq', OtherSequence()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('add0', IntegerType()),
        NamedType('add1', BooleanType()),
    )
    componentType = rootComponent + extensionAddition





import pytest
from asn1PERser.codec.per.encoder import encode as per_encoder
from asn1PERser.codec.per.encoder import SizeConstrainViolated, InvalidComponentIndexIntoStructuredType
from asn1PERser.classes.data.builtin.SequenceOfType import SequenceOfType
from asn1PERser.classes.data.builtin.IntegerType import IntegerType
from asn1PERser.classes.types.constraint import SequenceOfValueSize


I0_VAL = 34567
I1_VAL = -845
I2_VAL = 23
I3_VAL = 200000
I4_VAL = 5


def SCHEMA_no_constrains_sequence_of():
    class MySeqOf(SequenceOfType):
        '''
        MySeqOf::= SEQUENCE OF INTEGER
        '''
        componentType = IntegerType()

    return MySeqOf()


def SCHEMA_constrained_seq_of_no_extension(lb, ub):
    class MySeqOf(SequenceOfType):
        '''
        MySeqOf::= SEQUENCE (SIZE(lowerEndpoint..upperEndpoint)) OF INTEGER
        '''
        subtypeSpec = SequenceOfValueSize(lb, ub)
        componentType = IntegerType()

    return MySeqOf()


def SCHEMA_constrained_seq_of_extension_present(lb, ub):
    class MySeqOf(SequenceOfType):
        '''
        MySeqOf::= SEQUENCE (SIZE(lowerEndpoint..upperEndpoint,...)) OF INTEGER
        '''
        subtypeSpec = SequenceOfValueSize(lb, ub, extensionMarker=True)
        componentType = IntegerType()

    return MySeqOf()


def DATA_seq_of(schema_seq_of, i0_is, i1_is, i2_is, i3_is, i4_is):
    seq_of = schema_seq_of
    if i0_is:
        seq_of.extend([IntegerType(I0_VAL)])
    if i1_is:
        seq_of.extend([IntegerType(I1_VAL)])
    if i2_is:
        seq_of.extend([IntegerType(I2_VAL)])
    if i3_is:
        seq_of.extend([IntegerType(I3_VAL)])
    if i4_is:
        seq_of.extend([IntegerType(I4_VAL)])
    return seq_of


@pytest.mark.parametrize("sequence_of, encoded", [
    (DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=False, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '00'),
    (DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '0103008707'),
    (DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '020300870702FCB3'),
    (DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     '030300870702FCB30117'),
    (DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False),
     '040300870702FCB3011703030D40'),
    (DATA_seq_of(SCHEMA_no_constrains_sequence_of(), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     '050300870702FCB3011703030D400105'),
])
def test_no_constrains_sequence_of_integer_can_be_encoded(sequence_of, encoded):
    assert per_encoder(sequence_of) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence_of, encoded", [
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=0, ub=2), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '4003008707'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=0, ub=2), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '800300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '000300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     '400300870702FCB30117'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False),
     '800300870702FCB3011703030D40'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     'C00300870702FCB3011703030D400105'),
])
def test_constrained_sequence_of_no_extension_can_be_encoded(sequence_of, encoded):
    assert per_encoder(sequence_of) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence_of, encoded", [
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=1, ub=1), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '03008707'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=2, ub=2), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '0300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=3, ub=3), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     '0300870702FCB30117'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=4, ub=4), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False),
     '0300870702FCB3011703030D40'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=5, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     '0300870702FCB3011703030D400105'),
])
def test_constrained_sequence_of_of_fixed_length_no_extension_can_be_encoded(sequence_of, encoded):
    assert per_encoder(sequence_of) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence_of, encoded", [
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=2), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '2003008707'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=2), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '400300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '000300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     '200300870702FCB30117'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False),
     '400300870702FCB3011703030D40'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=2, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     '600300870702FCB3011703030D400105'),
])
def test_constrained_sequence_of_with_extension_and_num_of_elems_is_within_extension_root_can_be_encoded(sequence_of, encoded):
    assert per_encoder(sequence_of) == bytearray.fromhex(encoded)


@pytest.mark.parametrize("sequence_of, encoded", [
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '800103008707'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '80020300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=0, ub=0), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     '80030300870702FCB30117'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=1), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     '80020300870702FCB3'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=1), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     '80050300870702FCB3011703030D400105'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=3), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False),
     '80040300870702FCB3011703030D40'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=1, ub=3), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     '80050300870702FCB3011703030D400105'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=3, ub=4), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     '80050300870702FCB3011703030D400105'),
    (DATA_seq_of(SCHEMA_constrained_seq_of_extension_present(lb=3, ub=4), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     '800103008707'),
])
def test_constrained_sequence_of_with_extension_and_num_of_elems_is_not_within_extension_root_can_be_encoded(sequence_of, encoded):
    assert per_encoder(sequence_of) == bytearray.fromhex(encoded)

@pytest.mark.parametrize("sequence_of, exception", [
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=4, ub=5), i0_is=False, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     pytest.raises(SizeConstrainViolated)),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=4, ub=5), i0_is=True, i1_is=False, i2_is=False, i3_is=False, i4_is=False),
     pytest.raises(SizeConstrainViolated)),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=4, ub=5), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     pytest.raises(SizeConstrainViolated)),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=4, ub=5), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     pytest.raises(SizeConstrainViolated)),
])
def test_constrained_sequence_of_with_num_of_elems_less_than_lower_bound_and_no_extension_raises(sequence_of, exception):
    with exception:
        per_encoder(sequence_of)


@pytest.mark.parametrize("sequence_of, exception", [
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=1, ub=2), i0_is=True, i1_is=True, i2_is=True, i3_is=False, i4_is=False),
     pytest.raises(InvalidComponentIndexIntoStructuredType)),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=1, ub=2), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=False),
     pytest.raises(InvalidComponentIndexIntoStructuredType)),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=1, ub=2), i0_is=True, i1_is=True, i2_is=True, i3_is=True, i4_is=True),
     pytest.raises(InvalidComponentIndexIntoStructuredType)),
    (DATA_seq_of(SCHEMA_constrained_seq_of_no_extension(lb=1, ub=1), i0_is=True, i1_is=True, i2_is=False, i3_is=False, i4_is=False),
     pytest.raises(InvalidComponentIndexIntoStructuredType)),
])
def test_constrained_sequence_with_num_of_elems_greater_than_upper_bound_and_no_extension_raises(sequence_of, exception):
    with exception:
        per_encoder(sequence_of)


# -*- coding: utf-8 -*-
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="asn1PERser",
    version="0.4.0",
    author="Maciej Pikuła",
    author_email="erupikus@gmail.com",
    description="Parse ASN.1 schemas into Python code and encode/decode them using PER encoder",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/erupikus/asn1PERser",
    keywords = "asn asn1 asn.1 PER decoder encoder",
    packages=setuptools.find_packages(),
    package_data={
        '': ['*.txt']
    },
    install_requires=[
        "pyasn1>=0.4.7",
        "pyparsing>=2.4.7",
        "Jinja2>=2.11.3",
        "future>=0.18.2",
    ],
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import logging

import pytest
from pyasn1.type.namedtype import NamedType, OptionalNamedType
from asn1PERser import decode
from asn1PERser.classes.data.builtin import IntegerType, OctetStringType, BitStringType, BooleanType, \
    SequenceOfType, SequenceType, EnumeratedType, ChoiceType
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.test.parsing.asn1_python_code.Choice_Simple import Order, MyInteger, MyBoolean, MyBitString, \
    MyOctetString
from asn1PERser.test.parsing.asn1_python_code.SimpleTypes import MyEnumerated
from asn1PERser.test.parsing.asn1_python_code.SimpleProtocol import SimpleMessage, Port, UTC_Timestamp, \
    SequenceNumber, Start, Data, Payload, Message
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


@pytest.mark.parametrize("schema, value", [
    (IntegerType(10), 10),
    (OctetStringType(binValue='101001011100'), 'a5c0'),
    (OctetStringType(hexValue='DEADBEEF'), 'deadbeef'),
    (OctetStringType(hexValue='51800d80609bddfd80048054c0609bddfd800481de'), '51800d80609bddfd80048054c0609bddfd800481de'),
    (BitStringType(binValue='101001011100'), 2652),
    (BitStringType(hexValue='a5c0'), 42432),
    (BooleanType(True), True),
    (BooleanType(False), False),
])
def test_create_simple_types_dict(schema, value):
    assert schema.to_dict() == value


def create_choice(selection, initialized_type):
    order = Order()
    order[selection] = initialized_type
    return order


@pytest.mark.parametrize("schema, value", [
    (create_choice('one', IntegerType(12345)), {'one': 12345}),
    (create_choice('two', MyInteger(0)), {'two': 0}),
    (create_choice('three', OctetStringType(hexValue='feed')), {'three': 'feed'}),
    (create_choice('four', MyOctetString(hexValue='deefabcd')), {'four': 'deefabcd'}),
    (create_choice('five', BitStringType(binValue='111100001111')), {'five': 3855}),
    (create_choice('six', MyBitString(hexValue='babe')), {'six': 47806}),
    (create_choice('seven', BooleanType(True)), {'seven': True}),
    (create_choice('eight', MyBoolean(False)), {'eight': False}),
])
def test_create_choice_dict(schema, value):
    assert schema.to_dict() == value


@pytest.mark.parametrize("schema, value", [
    (MyEnumerated('one'), 'one'),
    (MyEnumerated('two'), 'two'),
    (MyEnumerated('three'), 'three'),
])
def test_create_enumerated_dict(schema, value):
    assert schema.to_dict() == value


def create_simple_sequence():
    class MySimpleSeq(SequenceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('one', IntegerType()),
            NamedType('two', BooleanType()),
            NamedType('three', MyEnumerated()),
            NamedType('four', BitStringType()),
            NamedType('five', OctetStringType()),
        )
        componentType = rootComponent

    mySeq = MySimpleSeq()
    mySeq['one'] = IntegerType(-1)
    mySeq['two'] = BooleanType(True)
    mySeq['three'] = MyEnumerated('one')
    mySeq['four'] = BitStringType(binValue='0111')
    mySeq['five'] = OctetStringType(hexValue='bada')

    return mySeq


def test_create_simple_sequence_dict():
    assert create_simple_sequence().to_dict() == {'one': -1,
                                                  'two': True,
                                                  'three': 'one',
                                                  'four': 7,
                                                  'five': 'bada'}


def create_simple_sequence_of(seq_of_type, values, input_type=None):
    class MySeqOf(SequenceOfType):
        componentType = seq_of_type()

    my_seq_of = MySeqOf()
    for value in values:
        if not input_type:
            my_seq_of.extend([seq_of_type(value)])
        elif input_type == 'hexValue':
            my_seq_of.extend([seq_of_type(hexValue=value)])
        elif input_type == 'binValue':
            my_seq_of.extend([seq_of_type(binValue=value)])

    return my_seq_of


@pytest.mark.parametrize("schema, value", [
    (create_simple_sequence_of(IntegerType, [1, 22, 333, 4444]), [1, 22, 333, 4444]),
    (create_simple_sequence_of(BitStringType, ['0011', '10', '10001000'], 'binValue'), [3, 2, 136]),
    (create_simple_sequence_of(OctetStringType, ['dead', 'beef'], 'hexValue'), ['dead', 'beef']),
])
def test_create_simple_sequence_of_dict(schema, value):
    assert schema.to_dict() == value


def test_simple_protocol_start_dict():
    src_port = Port(10000)
    dst_port = Port(20000)

    timestamp = UTC_Timestamp()
    timestamp['seconds'] = UTC_Timestamp.seconds(1234567)
    timestamp['useconds'] = UTC_Timestamp.useconds(7654321)

    sequence_num = SequenceNumber(1)

    start_message = Start()
    start_message['sequenceNumber'] = sequence_num
    start_message['timestamp'] = timestamp
    start_message['srcPort'] = src_port
    start_message['dstPort'] = dst_port

    simple_message = SimpleMessage()
    simple_message['start'] = start_message

    simple_message_dict = {'start': {'sequenceNumber': 1,
                                     'timestamp': {'seconds': 1234567,
                                                   'useconds': 7654321},
                                     'srcPort': 10000,
                                     'dstPort': 20000}}

    assert simple_message.to_dict() == simple_message_dict


def test_simple_protocol_data_dict():
    data_payload = Payload()
    data_payload.extend([Message(hexValue='DEAD')])
    data_payload.extend([Message(hexValue='BEEF')])
    data_payload.extend([Message(hexValue='FEED')])
    data_payload.extend([Message(hexValue='AA')])
    data_payload.extend([Message(hexValue='BBBBBBBB')])

    msg_data = Data()
    msg_data['sequenceNumber'] = SequenceNumber(55555)
    msg_data['swRelease'] = Data.swRelease('rel2')
    msg_data['macroId'] = Data.macroId(binValue='11110000111100001111')
    msg_data['payload'] = data_payload

    simple_message = SimpleMessage()
    simple_message['data'] = msg_data

    simple_message_dict = {'data': {'sequenceNumber': 55555,
                                    'swRelease': 'rel2',
                                    'macroId': 986895,
                                    'payload': ['dead', 'beef', 'feed', 'aa', 'bbbbbbbb']}}

    assert simple_message.to_dict() == simple_message_dict


def test_more_complicated_sequence():
    class FirstInnerSeq(SequenceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('aa', IntegerType()),
            NamedType('bb', IntegerType()),
        )
        componentType = rootComponent

    class SecondInnerSeq(SequenceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('cc', IntegerType()),
            NamedType('dd', IntegerType()),
        )
        componentType = rootComponent

    class FirstSeqOf(SequenceOfType):
        componentType = FirstInnerSeq()

    class SecondSeqOf(SequenceOfType):
        componentType = SecondInnerSeq()

    class TopSequence(SequenceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('one', FirstSeqOf()),
            NamedType('two', SecondSeqOf()),
        )
        componentType = rootComponent

    first_inner_seq_1 = FirstInnerSeq()
    first_inner_seq_1['aa'] = IntegerType(1)
    first_inner_seq_1['bb'] = IntegerType(2)

    first_inner_seq_2 = FirstInnerSeq()
    first_inner_seq_2['aa'] = IntegerType(3)
    first_inner_seq_2['bb'] = IntegerType(4)

    second_inner_seq_1 = SecondInnerSeq()
    second_inner_seq_1['cc'] = IntegerType(5)
    second_inner_seq_1['dd'] = IntegerType(6)

    second_inner_seq_2 = SecondInnerSeq()
    second_inner_seq_2['cc'] = IntegerType(7)
    second_inner_seq_2['dd'] = IntegerType(8)

    second_inner_seq_3 = SecondInnerSeq()
    second_inner_seq_3['cc'] = IntegerType(9)
    second_inner_seq_3['dd'] = IntegerType(10)

    first_seq_of = FirstSeqOf()
    first_seq_of.extend([first_inner_seq_1, first_inner_seq_2])

    second_seq_of = SecondSeqOf()
    second_seq_of.extend([second_inner_seq_1, second_inner_seq_2, second_inner_seq_3])

    top_seq = TopSequence()
    top_seq['one'] = first_seq_of
    top_seq['two'] = second_seq_of

    top_seq_dict = {'one': [{'aa': 1, 'bb': 2},
                            {'aa': 3, 'bb': 4}],
                    'two': [{'cc': 5, 'dd': 6},
                            {'cc': 7, 'dd': 8},
                            {'cc': 9, 'dd': 10}]}

    assert top_seq.to_dict() == top_seq_dict


def test_sequence_with_optional_types_dict():
    class Inner(SequenceType):
        rootComponent = AdditiveNamedTypes(
            OptionalNamedType('a', IntegerType()),
            OptionalNamedType('b', IntegerType()),
        )
        componentType = rootComponent

    class TopSeq(SequenceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('one', IntegerType()),
            OptionalNamedType('two', Inner()),
        )
        componentType = rootComponent

    top = TopSeq()
    top['one'] = IntegerType(0)

    assert top.to_dict() == {'one': 0}


@pytest.fixture
def sequence_with_extension_addition_group():
    class Port(IntegerType):
        subtypeSpec = ValueRange(10000, 65535)

    class SequenceNumber(IntegerType):
        subtypeSpec = ValueRange(0, 65535)

    class Start(SequenceType):
        subtypeSpec = ExtensionMarker(True)
        rootComponent = AdditiveNamedTypes(
            NamedType('sequenceNumber', SequenceNumber()),
        )
        extensionAdditionGroups = [
            AdditiveNamedTypes(
                NamedType('port', Port()),
            ),
        ]
        componentType = rootComponent + extensionAdditionGroups

    class Stop(SequenceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('sequenceNumber', SequenceNumber()),
        )
        componentType = rootComponent

    class SimpleMessage(ChoiceType):
        rootComponent = AdditiveNamedTypes(
            NamedType('start', Start()),
            NamedType('stop', Stop()),
        )
        componentType = rootComponent

    return SimpleMessage


def test_sequence_that_has_extension_addition_group_present(sequence_with_extension_addition_group):
    start_real_bytes = bytearray.fromhex("0004d2")
    decoded_start = decode(asn1Spec=sequence_with_extension_addition_group(), per_stream=start_real_bytes)
    assert decoded_start.to_dict() == {'start': {'sequenceNumber': 1234}}


def test_sequence_that_has_extension_addition_group_not_present(sequence_with_extension_addition_group):
    stop_real_bytes = bytearray.fromhex("8004d2")
    decoded_stop = decode(asn1Spec=sequence_with_extension_addition_group(), per_stream=stop_real_bytes)
    assert decoded_stop.to_dict() == {'stop': {'sequenceNumber': 1234}}



from ..type import ComponentType
from asn1PERser.classes.templates.creator import template_filler
from asn1PERser.classes.module import already_filled_template


class SequenceOfType(ComponentType):
    def __init__(self):
        super(SequenceOfType, self).__init__()
        self._ComponentType = None
        self.typereference = self.__class__.__name__

    def fill_template(self, has_parent=False):
        filled_template = template_filler.fill(asn_type=self.__class__.__name__,
                                               class_name=self.template_class_name,
                                               class_type=self.__class__.__name__,
                                               field_type=self.ComponentType.template_field_type,
                                               constraint={'extensionMarker': self.constraint.extensionMarker,
                                                           'lowerEndpoint': self.constraint.lowerEndpoint,
                                                           'upperEndpoint': self.constraint.upperEndpoint})
        if has_parent:
            return ([self.ComponentType], filled_template)
        if filled_template in already_filled_template:
            return ''
        already_filled_template.add(filled_template)
        if self.ComponentType.typereference in ['IntegerType', 'BooleanType', 'OctetStringType', 'BitStringType']:
            return filled_template
        components_template = self.ComponentType.fill_template()
        return components_template + filled_template

    @property
    def ComponentType(self):
        return self._ComponentType

    @ComponentType.setter
    def ComponentType(self, ComponentType):
        self._ComponentType = ComponentType

    def __getitem__(self, item):
        return [self.ComponentType][item]

    def __repr__(self):
        return '\n\t'.join([super(SequenceOfType, self).__repr__(), str(self.ComponentType)])



from asn1PERser.classes.parse_actions import parse_TypeAssignment, parse_ValueAssignment
from .lexical_items import valuereference, typereference, assignment
from .type_def import Type
from .value_def import Value


# 16.2
ValueAssignment = (valuereference + Type + assignment + Value).setParseAction(parse_ValueAssignment)


# 16.1
TypeAssignment = (typereference + assignment + Type).setParseAction(parse_TypeAssignment)



from pyasn1.type.constraint import ValueRangeConstraint, ConstraintsIntersection, ConstraintsUnion, SingleValueConstraint, \
    ValueSizeConstraint


MIN = float('-inf')
MAX = float('inf')


class NoConstraint(ConstraintsIntersection):
    def __init__(self):
        self.extensionMarker = False
        self.lowerEndpoint = None
        self.upperEndpoint = None
        super(NoConstraint, self).__init__()

    def __repr__(self):
        return str(None)


class ExtensionMarker(ConstraintsIntersection):
    def __init__(self, present=False):
        self.extensionMarker = present
        self.lowerEndpoint = None
        self.upperEndpoint = None
        super(ExtensionMarker, self).__init__()


class SequenceOfValueSize(ConstraintsIntersection):
    '''
    Use this constraint with SEQUENCE OF:
    with asn1 schema:

        MySeqOf::= SEQUENCE (SIZE(1..2)) OF OCTET STRING

    Python code:

        class MySeqOf(SequenceOfType):
            subtypeSpec SequenceOfValueSize(1, 2)
            componentType = OctetStringType()

    This is due to asn1PERser 'ValueSize' and pyasn1 'ValueSizeConstraint' will try to apply
    this size constraint check on componentType and NOT on SEQUENCE OF itself.
    The consequence of this is there will be no exception raised when MySeqOf is added 3rd (4th, 5th etc.)
    OCTET STRING.
    '''
    def __init__(self, lower_endpoint, upper_endpoint, extensionMarker=False):
        self.extensionMarker = extensionMarker
        self.lowerEndpoint = None if lower_endpoint == MIN else lower_endpoint
        self.upperEndpoint = None if upper_endpoint == MAX else upper_endpoint
        super(SequenceOfValueSize, self).__init__()


class ValueRange(ValueRangeConstraint):
    def __init__(self, lower_endpoint, upper_endpoint, extensionMarker=False):
        self.extensionMarker = extensionMarker
        self.lowerEndpoint = None if lower_endpoint == MIN else lower_endpoint
        self.upperEndpoint = None if upper_endpoint == MAX else upper_endpoint
        super(ValueRange, self).__init__(lower_endpoint, upper_endpoint)

    def _testValue(self, value, idx):
        if not self.extensionMarker:
            return super(ValueRange, self)._testValue(value, idx)


class SingleValue(SingleValueConstraint):
    def __init__(self, extension_marker=False, *values):
        self.extensionMarker = extension_marker
        self.lowerEndpoint = min(values)
        self.upperEndpoint = max(values)
        super(SingleValue, self).__init__(*values)

    def _testValue(self, value, idx):
        if not self.extensionMarker:
            return super(SingleValue, self)._testValue(value, idx)


class ValueSize(ValueSizeConstraint):
    def __init__(self, lower_endpoint, upper_endpoint, extensionMarker=False):
        self.extensionMarker = extensionMarker
        self.lowerEndpoint = None if lower_endpoint == MIN else lower_endpoint
        self.upperEndpoint = None if upper_endpoint == MAX else upper_endpoint
        super(ValueSize, self).__init__(lower_endpoint, upper_endpoint)

    def _testValue(self, value, idx):
        if not self.extensionMarker:
            return super(ValueSize, self)._testValue(value, idx)


class ConstraintOr(ConstraintsUnion):
    def __init__(self, extensionMarker=False, *constraints):
        self.extensionMarker = extensionMarker
        self.lowerEndpoint = min([constraint.lowerEndpoint for constraint in constraints])
        self.upperEndpoint = max([constraint.upperEndpoint for constraint in constraints])
        super(ConstraintOr, self).__init__(*constraints)

    def _testValue(self, value, idx):
        if not self.extensionMarker:
            return super(ConstraintOr, self)._testValue(value, idx)


class ConstraintAnd(ConstraintsIntersection):
    def __init__(self, extensionMarker=False, *constraints):
        self.extensionMarker = extensionMarker
        self.lowerEndpoint = min([constraint.lowerEndpoint for constraint in constraints])
        self.upperEndpoint = max([constraint.upperEndpoint for constraint in constraints])
        super(ConstraintAnd, self).__init__(*constraints)

    def _testValue(self, value, idx):
        if not self.extensionMarker:
            return super(ConstraintAnd, self)._testValue(value, idx)



from .BitStringType import BitStringType
from .BooleanType import BooleanType
from .ChoiceType import ChoiceType
from .DateType import DateType
from .DefinedType import DefinedType
from .EnumeratedType import EnumeratedType
from .IA5String import IA5String
from .IntegerType import IntegerType
from .NullType import NullType
from .NumericString import NumericString
from .OctetStringType import OctetStringType
from .RealType import RealType
from .SequenceType import SequenceType
from .SequenceOfType import SequenceOfType
from .UTF8String import UTF8String
from .VisibleString import VisibleString


builtin_types = ['BitStringType', 'BooleanType', 'ChoiceType', 'DateType', 'EnumeratedType', 'IA5String', 'IntegerType',
                 'NullType', 'NumericString', 'OctetStringType', 'RealType', 'SequenceType', 'SequenceOfType',
                 'UTF8String', 'VisibleString']



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class MyEnum(EnumeratedType):
    enumerationRoot = NamedValues(
        ('one', 0),
        ('two', 1),
    )
    namedValues = enumerationRoot


class MySeq(SequenceType):
    class i0(IntegerType):
        subtypeSpec = ValueRange(10, 20)

    rootComponent = AdditiveNamedTypes(
        NamedType('i0', i0()),
        NamedType('i1', IntegerType()),
    )
    componentType = rootComponent


class Data1(SequenceType):
    class d1(BitStringType):
        subtypeSpec = ValueSize(1, 20, extensionMarker=True)

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('d0', OctetStringType()),
    )
    extensionAdditionGroups = [
        AdditiveNamedTypes(
            NamedType('d1', d1()),
            NamedType('d2', MySeq()),
        ),
    ]
    componentType = rootComponent + extensionAdditionGroups


class Data2(ChoiceType):
    class c0(IntegerType):
        subtypeSpec = ValueRange(1, MAX)

    class c3(OctetStringType):
        subtypeSpec = ValueSize(5, 10)

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('c0', c0()),
    )
    extensionAdditionGroups = [
        AdditiveNamedTypes(
            NamedType('c1', MyEnum()),
            NamedType('c2', BooleanType()),
        ),
        AdditiveNamedTypes(
            NamedType('c3', c3()),
            NamedType('c4', MySeq()),
        ),
    ]
    componentType = rootComponent + extensionAdditionGroups


class Data3(SequenceType):
    class s3(IntegerType):
        subtypeSpec = ValueRange(-10, 100)

    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('s0', IntegerType()),
    )
    extensionAddition = AdditiveNamedTypes(
        NamedType('s4', IntegerType()),
        NamedType('s5', IntegerType()),
    )
    extensionAdditionGroups = [
        AdditiveNamedTypes(
            NamedType('s1', MySeq()),
            OptionalNamedType('s2', IntegerType()),
        ),
        AdditiveNamedTypes(
            NamedType('s3', s3()),
            NamedType('s6', MyEnum()),
        ),
    ]
    componentType = rootComponent + extensionAddition + extensionAdditionGroups





from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class MyInteger(IntegerType):
    pass


class MyOctetString(OctetStringType):
    pass


class MyBitString(BitStringType):
    pass


class MyBoolean(BooleanType):
    pass


class MyEnumerated(EnumeratedType):
    enumerationRoot = NamedValues(
        ('four', 0),
        ('five', 1),
        ('six', 2),
    )
    namedValues = enumerationRoot


class MySeq(SequenceType):
    class nine(EnumeratedType):
        enumerationRoot = NamedValues(
            ('one', 0),
            ('two', 1),
            ('three', 2),
        )
        namedValues = enumerationRoot

    rootComponent = AdditiveNamedTypes(
        NamedType('one', IntegerType()),
        NamedType('two', MyInteger()),
        NamedType('three', OctetStringType()),
        NamedType('four', MyOctetString()),
        NamedType('five', BitStringType()),
        NamedType('six', MyBitString()),
        NamedType('seven', BooleanType()),
        NamedType('eight', MyBoolean()),
        NamedType('nine', nine()),
        NamedType('ten', MyEnumerated()),
    )
    componentType = rootComponent





from pyasn1.type.univ import SequenceOf, noValue
from asn1PERser.codec.per.encoder import encode_sequence_of
from asn1PERser.codec.per.decoder import decode_sequence_of
from asn1PERser.classes.types.constraint import SequenceOfValueSize, MAX


class SequenceOfType(SequenceOf):
    subtypeSpec = SequenceOfValueSize(0, MAX)

    def fill_field_list(self, field_list):
        sequence_of_field_list = encode_sequence_of(self)
        if sequence_of_field_list:
            field_list.extend(sequence_of_field_list)

    def create_field_list(self, per_bytes):
        decoded = decode_sequence_of(self, per_bytes)
        return decoded

    def to_dict(self, is_root=True):
        component_list = []
        for componentValue in self:
            if (componentValue is not noValue and componentValue.isValue) or hasattr(componentValue, 'componentType'):
                value_dict = componentValue.to_dict(is_root=False)
                if value_dict is not noValue:
                    component_list.append(value_dict)

        if is_root:
            return component_list
        if not component_list:
            return noValue
        return component_list



import pytest
from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd
from asn1PERser.codec.per.encoder import encode as per_encoder


#-------------------------------------
class DataSeq(SequenceType):
    rootComponent = NamedTypes(
        NamedType('myBool0', BooleanType()),
        NamedType('myBitString', BitStringType())
    )
    componentType = rootComponent


class some_enum(EnumeratedType):
    namedValues = NamedValues(
        ('one', 0),
        ('two', 1),
        ('three', 2)
    )


class some_seqOf(SequenceOfType):
    componentType = OctetStringType()


class MySeq(SequenceType):
    rootComponent = NamedTypes(
        NamedType('some-seq', DataSeq()),
        NamedType('some-enum', some_enum()),
        NamedType('some-integer', IntegerType()),
        NamedType('some-seqOf', some_seqOf())
    )
    componentType = rootComponent


def DATA_first_sequence():
    '''
    mySeq MySeq ::= {
        some-seq {
            myBool0    TRUE,
            myBitString    '1111'B
        },
        some-enum two,
        some-integer  100,
        some-seqOf {'DEAD'H, 'BEEF'H}
    }
    '''
    data_seq = DataSeq()
    data_seq['myBool0'] = BooleanType(True)
    data_seq['myBitString'] = BitStringType(binValue='1111')

    some_seqOf_component = some_seqOf()
    some_seqOf_component.extend([OctetStringType(hexValue='DEAD'),
                                 OctetStringType(hexValue='BEEF')])

    my_seq = MySeq()
    my_seq['some-seq'] = data_seq
    my_seq['some-enum'] = some_enum('two')
    my_seq['some-integer'] = IntegerType(100)
    my_seq['some-seqOf'] = some_seqOf_component

    return my_seq


def test_first_sequence_can_be_encoded():
    per_encoded = per_encoder(asn1Spec=DATA_first_sequence())
    assert per_encoded == bytearray.fromhex('8004F401640202DEAD02BEEF')

#-------------------------------------


'''
OuterSeq ::= SEQUENCE {
    sof1   MySof1 OPTIONAL,
    sof2   MySof2 OPTIONAL
}

MySof1 ::= SEQUENCE (SIZE(1..2)) OF InnerSeq1
MySof2 ::= SEQUENCE (SIZE(1..2)) OF InnerSeq2

InnerSeq1 ::= SEQUENCE {
    i0   INTEGER,
    i1   INTEGER
}

InnerSeq2 ::= SEQUENCE {
    i2   INTEGER,
    i3   INTEGER
}
'''

class InnerSeq1(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('i0', IntegerType()),
        NamedType('i1', IntegerType()),
    )
    componentType = rootComponent


class InnerSeq2(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('i2', IntegerType()),
        NamedType('i3', IntegerType()),
    )
    componentType = rootComponent


class MySof1(SequenceOfType):
    subtypeSpec = SequenceOfValueSize(1, 2)
    componentType = InnerSeq1()


class MySof2(SequenceOfType):
    subtypeSpec = SequenceOfValueSize(1, 2)
    componentType = InnerSeq2()


class OuterSeq(SequenceType):
    rootComponent = AdditiveNamedTypes(
        OptionalNamedType('sof1', MySof1()),
        OptionalNamedType('sof2', MySof2()),
    )
    componentType = rootComponent


def DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=False, my_sof2_is=False):
    inner_seq_1 = InnerSeq1()
    inner_seq_1['i0'] = IntegerType(10)
    inner_seq_1['i1'] = IntegerType(20)

    inner_seq_2 = InnerSeq2()
    inner_seq_2['i2'] = IntegerType(30)
    inner_seq_2['i3'] = IntegerType(40)

    my_sof1 = MySof1()
    my_sof1.extend([inner_seq_1])

    my_sof2 = MySof2()
    my_sof2.extend([inner_seq_2])

    outer = OuterSeq()

    if my_sof1_is:
        outer['sof1'] = my_sof1
    if my_sof2_is:
        outer['sof2'] = my_sof2
    return outer


@pytest.mark.parametrize("data, encoded", [
    (DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=True, my_sof2_is=False), '80010A0114'),
    (DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=True, my_sof2_is=True), 'C0010A011400011E0128'),
    (DATA_seq__of__seq_of__of__seq_in_root(my_sof1_is=False, my_sof2_is=True), '40011E0128'),
])
def test_seq_of_seq_of_of_seq_in_root_can_be_encoded(data, encoded):
    assert per_encoder(asn1Spec=data) == bytearray.fromhex(encoded)


#-----------------------------


'''
OuterSeq ::= SEQUENCE {
    s0     INTEGER,
    ...,
    sof1   MySof1 OPTIONAL,
    sof2   MySof2 OPTIONAL
}

MySof1 ::= SEQUENCE (SIZE(1..2)) OF InnerSeq1
MySof2 ::= SEQUENCE (SIZE(1..2)) OF InnerSeq2

InnerSeq1 ::= SEQUENCE {
    i0   INTEGER,
    i1   INTEGER
}

InnerSeq2 ::= SEQUENCE {
    i2   INTEGER,
    i3   INTEGER
}
'''
class OuterSeq1(SequenceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('s0', IntegerType())
    )
    extensionAddition = AdditiveNamedTypes(
        OptionalNamedType('sof1', MySof1()),
        OptionalNamedType('sof2', MySof2()),
    )
    componentType = rootComponent + extensionAddition


def DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=False, my_sof2_is=False):
    inner_seq_1 = InnerSeq1()
    inner_seq_1['i0'] = IntegerType(10)
    inner_seq_1['i1'] = IntegerType(20)

    inner_seq_2 = InnerSeq2()
    inner_seq_2['i2'] = IntegerType(30)
    inner_seq_2['i3'] = IntegerType(40)

    my_sof1 = MySof1()
    my_sof1.extend([inner_seq_1])

    my_sof2 = MySof2()
    my_sof2.extend([inner_seq_2])

    outer = OuterSeq1()

    outer['s0'] = IntegerType(-10)
    if my_sof1_is:
        outer['sof1'] = my_sof1
    if my_sof2_is:
        outer['sof2'] = my_sof2
    return outer


@pytest.mark.parametrize("data, encoded", [
    (DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=True, my_sof2_is=False), '8001F603000500010A0114'),
    (DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=True, my_sof2_is=True), '8001F603800500010A01140500011E0128'),
    (DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=False, my_sof2_is=True), '8001F602800500011E0128'),
    (DATA_seq__of__seq_of__of__seq_in_ext_add(my_sof1_is=False, my_sof2_is=False), '0001F6'),
])
def test_seq_of_seq_of_of_seq_in_extension_addition_can_be_encoded(data, encoded):
    assert per_encoder(asn1Spec=data) == bytearray.fromhex(encoded)

#----------------------------------


'''
OuterSeq ::= SEQUENCE {
    s0     INTEGER,
    ...,
    [[
    sof1   MySof1 OPTIONAL,
    sof2   MySof2 OPTIONAL
    ]]
}

MySof1 ::= SEQUENCE (SIZE(1..2)) OF InnerSeq1
MySof2 ::= SEQUENCE (SIZE(1..2)) OF InnerSeq2

InnerSeq1 ::= SEQUENCE {
    i0   INTEGER,
    i1   INTEGER
}

InnerSeq2 ::= SEQUENCE {
    i2   INTEGER,
    i3   INTEGER
}
'''
class OuterSeq2(SequenceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('s0', IntegerType())
    )
    extensionAdditionGroups = [
        AdditiveNamedTypes(
            OptionalNamedType('sof1', MySof1()),
            OptionalNamedType('sof2', MySof2()),
        )
    ]
    componentType = rootComponent + extensionAdditionGroups


def DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=False, my_sof2_is=False):
    inner_seq_1 = InnerSeq1()
    inner_seq_1['i0'] = IntegerType(10)
    inner_seq_1['i1'] = IntegerType(20)

    inner_seq_2 = InnerSeq2()
    inner_seq_2['i2'] = IntegerType(30)
    inner_seq_2['i3'] = IntegerType(40)

    my_sof1 = MySof1()
    my_sof1.extend([inner_seq_1])

    my_sof2 = MySof2()
    my_sof2.extend([inner_seq_2])

    outer = OuterSeq2()

    outer['s0'] = IntegerType(-10)
    if my_sof1_is:
        outer['sof1'] = my_sof1
    if my_sof2_is:
        outer['sof2'] = my_sof2
    return outer


@pytest.mark.parametrize("data, encoded", [
    (DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=True, my_sof2_is=False), '8001F6010580010A0114'),
    (DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=True, my_sof2_is=True), '8001F6010AC0010A011400011E0128'),
    (DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=False, my_sof2_is=True), '8001F6010540011E0128'),
    (DATA_seq__of__seq_of__of__seq_in_ext_add_group(my_sof1_is=False, my_sof2_is=False), '0001F6'),
])
def test_seq_of_seq_of_of_seq_in_extension_addition_group_can_be_encoded(data, encoded):
    assert per_encoder(asn1Spec=data) == bytearray.fromhex(encoded)



import pytest
from asn1PERser.codec.per.decoder import decode as per_decoder
from asn1PERser.classes.data.builtin.IntegerType import IntegerType
from asn1PERser.classes.types.constraint import ValueRange, MIN, MAX


def SCHEMA_not_constrained_integer():
    class MyInt(IntegerType):
        '''
        MyInt ::= INTEGER
        '''
        pass
    return MyInt


def SCHEMA_constrained_integer(lowerEndpoint_value, upperEndpoint_value, extensionMarker_value=False):
    class MyInt(IntegerType):
        '''
        MyInt ::= INTEGER (lowerEndpoint..upperEndpoint, extensionMarker)
        '''
        subtypeSpec = ValueRange(lowerEndpoint_value, upperEndpoint_value, extensionMarker=extensionMarker_value)
    return MyInt


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_not_constrained_integer(), '0100', 0),
    (SCHEMA_not_constrained_integer(), '0101', 1),
    (SCHEMA_not_constrained_integer(), '0108', 8),
    (SCHEMA_not_constrained_integer(), '010F', 15),
    (SCHEMA_not_constrained_integer(), '0110', 16),
    (SCHEMA_not_constrained_integer(), '0140', 64),
    (SCHEMA_not_constrained_integer(), '017F', 127),
    (SCHEMA_not_constrained_integer(), '020080', 128),
    (SCHEMA_not_constrained_integer(), '020081', 129),
    (SCHEMA_not_constrained_integer(), '020FA0', 4000),
    (SCHEMA_not_constrained_integer(), '027FFF', 32767),
    (SCHEMA_not_constrained_integer(), '03008000', 32768),
    (SCHEMA_not_constrained_integer(), '037FFFFF', 8388607),
    (SCHEMA_not_constrained_integer(), '0400800000', 8388608),
    (SCHEMA_not_constrained_integer(), '0401803E6F', 25181807),
])
def test_unconstrained_greater_or_equal_zero_integers_are_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_not_constrained_integer(), '04FE7FC191', -25181807),
    (SCHEMA_not_constrained_integer(), '03800000', -8388608),
    (SCHEMA_not_constrained_integer(), '03800001', -8388607),
    (SCHEMA_not_constrained_integer(), '028000', -32768),
    (SCHEMA_not_constrained_integer(), '028001', -32767),
    (SCHEMA_not_constrained_integer(), '02F060', -4000),
    (SCHEMA_not_constrained_integer(), '02FF7F', -129),
    (SCHEMA_not_constrained_integer(), '0180', -128),
    (SCHEMA_not_constrained_integer(), '0181', -127),
    (SCHEMA_not_constrained_integer(), '01C0', -64),
    (SCHEMA_not_constrained_integer(), '01F0', -16),
    (SCHEMA_not_constrained_integer(), '01F1', -15),
    (SCHEMA_not_constrained_integer(), '01F8', -8),
    (SCHEMA_not_constrained_integer(), '01FF', -1),
])
def test_unconstrained_lower_than_zero_integers_are_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-8388607, upperEndpoint_value=-8388607), '00', -8388607),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1864, upperEndpoint_value=-1864), '00', -1864),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-56), '00', -56),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-3, upperEndpoint_value=-3), '00', -3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=0), '00', 0),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1, upperEndpoint_value=1), '00', 1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8, upperEndpoint_value=8), '00', 8),
    (SCHEMA_constrained_integer(lowerEndpoint_value=15, upperEndpoint_value=15), '00', 15),
    (SCHEMA_constrained_integer(lowerEndpoint_value=16, upperEndpoint_value=16), '00', 16),
    (SCHEMA_constrained_integer(lowerEndpoint_value=64, upperEndpoint_value=64), '00', 64),
    (SCHEMA_constrained_integer(lowerEndpoint_value=127, upperEndpoint_value=127), '00', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=128, upperEndpoint_value=128), '00', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=129, upperEndpoint_value=129), '00', 129),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4000), '00', 4000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=32767, upperEndpoint_value=32767), '00', 32767),
    (SCHEMA_constrained_integer(lowerEndpoint_value=32768, upperEndpoint_value=32768), '00', 32768),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8388607, upperEndpoint_value=8388607), '00', 8388607),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8388608, upperEndpoint_value=8388608), '00', 8388608),
    (SCHEMA_constrained_integer(lowerEndpoint_value=25181807, upperEndpoint_value=25181807, ), '00', 25181807),
])
def test_constrained_integer_range_has_the_value_1_so_decoding_shall_be_empty(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-8388607, upperEndpoint_value=-8388606), '00', -8388607),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-8388607, upperEndpoint_value=-8388606), '80', -8388606),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1864, upperEndpoint_value=-1863), '00', -1864),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1864, upperEndpoint_value=-1863), '80', -1863),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-55), '00', -56),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-55), '80', -55),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-3, upperEndpoint_value=-2), '00', -3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-3, upperEndpoint_value=-2), '80', -2),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1, upperEndpoint_value=0), '00', -1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1, upperEndpoint_value=0), '80', 0),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1, upperEndpoint_value=2), '00', 1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1, upperEndpoint_value=2), '80', 2),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8, upperEndpoint_value=9), '00', 8),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8, upperEndpoint_value=9), '80', 9),
    (SCHEMA_constrained_integer(lowerEndpoint_value=15, upperEndpoint_value=16), '00', 15),
    (SCHEMA_constrained_integer(lowerEndpoint_value=15, upperEndpoint_value=16), '80', 16),
    (SCHEMA_constrained_integer(lowerEndpoint_value=16, upperEndpoint_value=17), '00', 16),
    (SCHEMA_constrained_integer(lowerEndpoint_value=16, upperEndpoint_value=17), '80', 17),
    (SCHEMA_constrained_integer(lowerEndpoint_value=64, upperEndpoint_value=65), '00', 64),
    (SCHEMA_constrained_integer(lowerEndpoint_value=64, upperEndpoint_value=65), '80', 65),
    (SCHEMA_constrained_integer(lowerEndpoint_value=127, upperEndpoint_value=128), '00', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=127, upperEndpoint_value=128), '80', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=128, upperEndpoint_value=129), '00', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=128, upperEndpoint_value=129), '80', 129),
    (SCHEMA_constrained_integer(lowerEndpoint_value=129, upperEndpoint_value=130), '00', 129),
    (SCHEMA_constrained_integer(lowerEndpoint_value=129, upperEndpoint_value=130), '80', 130),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4001), '00', 4000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4001), '80', 4001),
    (SCHEMA_constrained_integer(lowerEndpoint_value=32767, upperEndpoint_value=32768), '00', 32767),
    (SCHEMA_constrained_integer(lowerEndpoint_value=32767, upperEndpoint_value=32768), '80', 32768),
    (SCHEMA_constrained_integer(lowerEndpoint_value=32768, upperEndpoint_value=32769), '00', 32768),
    (SCHEMA_constrained_integer(lowerEndpoint_value=32768, upperEndpoint_value=32769), '80', 32769),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8388607, upperEndpoint_value=8388608), '00', 8388607),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8388607, upperEndpoint_value=8388608), '80', 8388608),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8388608, upperEndpoint_value=8388609), '00', 8388608),
    (SCHEMA_constrained_integer(lowerEndpoint_value=8388608, upperEndpoint_value=8388609), '80', 8388609),
    (SCHEMA_constrained_integer(lowerEndpoint_value=25181807, upperEndpoint_value=25181808, ), '00', 25181807),
    (SCHEMA_constrained_integer(lowerEndpoint_value=25181807, upperEndpoint_value=25181808, ), '80', 25181808),
])
def test_constrained_integer_range_has_value_2_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-8388610, upperEndpoint_value=-8388608), '00', -8388610),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-8388610, upperEndpoint_value=-8388608), '40', -8388609),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-8388610, upperEndpoint_value=-8388608), '80', -8388608),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1, upperEndpoint_value=1), '00', -1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1, upperEndpoint_value=1), '40', 0),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1, upperEndpoint_value=1), '80', 1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=25181807, upperEndpoint_value=25181809, ), '00', 25181807),
    (SCHEMA_constrained_integer(lowerEndpoint_value=25181807, upperEndpoint_value=25181809, ), '40', 25181808),
    (SCHEMA_constrained_integer(lowerEndpoint_value=25181807, upperEndpoint_value=25181809, ), '80', 25181809),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-53), '00', -56),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-53), '40', -55),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-53), '80', -54),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-56, upperEndpoint_value=-53), 'C0', -53),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4003), '00', 4000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4003), '40', 4001),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4003), '80', 4002),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4003), 'C0', 4003),
])
def test_constrained_integer_range_has_value_3_4_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=2), '00', -2),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=2), '20', -1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=2), '40', 0),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=2), '60', 1),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=2), '80', 2),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=3), 'A0', 3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=4), 'C0', 4),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-2, upperEndpoint_value=5), 'E0', 5),
])
def test_constrained_integer_range_has_value_5_6_7_8_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '00', -100000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '10', -99999),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '20', -99998),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '30', -99997),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '40', -99996),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '50', -99995),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '60', -99994),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '70', -99993),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99992), '80', -99992),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99991), '90', -99991),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99990), 'A0', -99990),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99989), 'B0', -99989),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99988), 'C0', -99988),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99987), 'D0', -99987),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99986), 'E0', -99986),
    (SCHEMA_constrained_integer(lowerEndpoint_value=-100000, upperEndpoint_value=-99985), 'F0', -99985),
])
def test_constrained_integer_range_has_value_9_to_16_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-4016, upperEndpoint_value=-4000), '78', -4001),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4016), '08', 4001),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4016), '18', 4003),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4016), '28', 4005),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4016), '50', 4010),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4016), '70', 4014),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4016), '80', 4016),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4017), '88', 4017),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4020), 'A0', 4020),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4024), 'C0', 4024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4028), 'E0', 4028),
    (SCHEMA_constrained_integer(lowerEndpoint_value=4000, upperEndpoint_value=4031), 'F8', 4031),
])
def test_constrained_integer_range_has_value_17_to_32_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1000032, upperEndpoint_value=-1000000), '7C', -1000001),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '04', 1000001),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '14', 1000005),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '24', 1000009),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '40', 1000016),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '58', 1000022),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '5C', 1000023),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '60', 1000024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '70', 1000028),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000032), '80', 1000032),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000040), '94', 1000037),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000045), 'B4', 1000045),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000056), 'DC', 1000055),
    (SCHEMA_constrained_integer(lowerEndpoint_value=1000000, upperEndpoint_value=1000063), 'FC', 1000063),
])
def test_constrained_integer_range_has_value_33_to_64_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-64, upperEndpoint_value=0), '72', -7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=64), '0E', 7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=64), '32', 25),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=64), '58', 44),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=64), '7A', 61),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=70), '8A', 69),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=77), '9A', 77),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=85), 'AA', 85),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=98), 'C4', 98),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=111), 'DE', 111),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=120), 'EE', 119),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=127), 'FE', 127),
])
def test_constrained_integer_range_has_value_65_to_128_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-128, upperEndpoint_value=0), '79', -7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '07', 7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '19', 25),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '2C', 44),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '3D', 61),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '45', 69),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '4D', 77),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '55', 85),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '62', 98),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '6F', 111),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '77', 119),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '7F', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=128), '80', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=140), '8C', 140),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=161), 'A1', 161),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=182), 'B6', 182),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=230), 'E6', 230),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=253), 'FD', 253),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=254), 'FE', 254),
])
def test_constrained_integer_range_has_value_129_to_255_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-255, upperEndpoint_value=0), 'F8', -7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '07', 7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '19', 25),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '2C', 44),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '3D', 61),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '45', 69),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '4D', 77),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '55', 85),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '62', 98),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '6F', 111),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '77', 119),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '7F', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '80', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), '8C', 140),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), 'A1', 161),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), 'B6', 182),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), 'E6', 230),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), 'FD', 253),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), 'FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255), 'FF', 255),
])
def test_constrained_integer_range_has_value_256_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-256, upperEndpoint_value=0), '00F9', -7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0007', 7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0019', 25),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '002C', 44),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '003D', 61),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0045', 69),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '004D', 77),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0055', 85),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0062', 98),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '006F', 111),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0077', 119),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '007F', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0080', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '008C', 140),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '00A1', 161),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '00B6', 182),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '00E6', 230),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '00FD', 253),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '00FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '00FF', 255),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=256), '0100', 256),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=511), '01FF', 511),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=512), '0200', 512),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1023), '03FF', 1023),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1024), '0400', 1024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1024), '0400', 1024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=8191), '1FFF', 8191),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=8192), '2000', 8192),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=32768), '7FFE', 32766),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=32768), '7FFF', 32767),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=32768), '8000', 32768),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=44321), 'AD21', 44321),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=60000), 'D903', 55555),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=65535), 'EA60', 60000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=65535), 'FFFE', 65534),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=65535), 'FFFF', 65535),
])
def test_constrained_integer_range_has_value_257_to_65536_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=-1234567, upperEndpoint_value=0), '80114FE7', -100000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0007', 7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0019', 25),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '002C', 44),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '003D', 61),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0045', 69),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '004D', 77),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0055', 85),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0062', 98),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '006F', 111),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0077', 119),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '007F', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '0080', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '008C', 140),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '00A1', 161),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '00B6', 182),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '00E6', 230),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '00FD', 253),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '00FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '00FF', 255),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '400100', 256),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '4001FF', 511),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '400200', 512),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '4003FF', 1023),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '400400', 1024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '400400', 1024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '401FFF', 8191),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '402000', 8192),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '407FFE', 32766),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '407FFF', 32767),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '408000', 32768),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '40AD21', 44321),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '40D903', 55555),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '40EA60', 60000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '40FFFE', 65534),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '40FFFF', 65535),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '80010000', 65536),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1234567), '800186A0', 100000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=256, upperEndpoint_value=1234567), '0000', 256),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), '0064', 100),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), '00FF', 255),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), '400100', 256),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), '800186A0', 100000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), 'C005F5E100', 100000000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), 'C0FFFFFFFE', 4294967294),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=4294967295), 'C0FFFFFFFF', 4294967295),
])
def test_constrained_integer_range_has_value_greater_than_65536_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0107', 7),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0119', 25),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '012C', 44),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '013D', 61),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0145', 69),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '014D', 77),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0155', 85),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0162', 98),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '016F', 111),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0177', 119),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '017F', 127),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0180', 128),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '018C', 140),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '01A1', 161),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '01B6', 182),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '01E6', 230),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '01FD', 253),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '01FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '01FF', 255),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '020100', 256),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0201FF', 511),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '020200', 512),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '0203FF', 1023),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '020400', 1024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '020400', 1024),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '021FFF', 8191),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '022000', 8192),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '027FFE', 32766),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '027FFF', 32767),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '028000', 32768),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '02AD21', 44321),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '02D903', 55555),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '02EA60', 60000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '02FFFE', 65534),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '02FFFF', 65535),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '03010000', 65536),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=MAX), '030186A0', 100000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=256, upperEndpoint_value=MAX), '0100', 256)
])
def test_semi_constrained_integer_is_encoded_can_be_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '60', 3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=120, extensionMarker_value=True), '03', 3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255, extensionMarker_value=True), '0003', 3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=10000, extensionMarker_value=True), '000003', 3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '0003', 3),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=255, extensionMarker_value=True), '00FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=10000, extensionMarker_value=True), '0000FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '00FE', 254),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=10000, extensionMarker_value=True), '0003FF', 1023),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '2003FF', 1023),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=100000, extensionMarker_value=True), '207FFE', 32766),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1000000, extensionMarker_value=True), '20FFFF', 65535),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=1000000, extensionMarker_value=True), '400186A0', 100000),
])
def test_extension_marker_is_present_and_value_is_within_extension_root_integer_is_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value


@pytest.mark.parametrize("schema, encoded, value", [
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '800104', 4),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '80030186A0', 100000),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '8001FC', -4),
    (SCHEMA_constrained_integer(lowerEndpoint_value=0, upperEndpoint_value=3, extensionMarker_value=True), '8003FE7960', -100000),
])
def test_extension_marker_is_present_and_values_in_not_within_extension_root_integer_is_decoded(schema, encoded, value):
    assert per_decoder(per_stream=bytearray.fromhex(encoded), asn1Spec=schema()) == value



import pytest


@pytest.mark.parametrize("asn1_schema, python_code", [
    ('SimpleTypes.asn1', 'SimpleTypes.py'),
    ('Types_constrained.asn1', 'Types_constrained.py'),
    ('Sequence_Simple.asn1', 'Sequence_Simple.py'),
    ('Sequence_Mixed.asn1', 'Sequence_Mixed.py'),
    ('Choice_Simple.asn1', 'Choice_Simple.py'),
    ('ComponentTypes.asn1', 'ComponentTypes.py'),
    ('ComponentTypes_extensionMarker.asn1', 'ComponentTypes_extensionMarker.py'),
    ('ComponentTypes_extensionAddition.asn1', 'ComponentTypes_extensionAddition.py'),
    ('Sequence_Default.asn1', 'Sequence_Default.py'),
    ('Types_containing_constrained_types.asn1', 'Types_containing_constrained_types.py'),
    ('ComponentTypes_containing_ComponentTypes.asn1', 'ComponentTypes_containing_ComponentTypes.py'),
    ('SimpleProtocol.asn1', 'SimpleProtocol.py'),
    ('ComponentTypes_extensionAdditionGroup.asn1', 'ComponentTypes_extensionAdditionGroup.py'),
    ('DefinedType_Simple.asn1', 'DefinedType_Simple.py'),
], indirect=['asn1_schema', 'python_code'])
def test_schema(asn1_schema, python_code, parse_schema):
    parsed_schema = parse_schema(asn1_schema)
    assert parsed_schema == python_code



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType, DefaultedNamedType
from pyasn1.type.namedval import NamedValues
from asn1PERser.classes.data.builtin import *
from asn1PERser.classes.types.type import AdditiveNamedTypes
from asn1PERser.classes.types.constraint import MIN, MAX, NoConstraint, ExtensionMarker, SequenceOfValueSize, \
    ValueRange, SingleValue, ValueSize, ConstraintOr, ConstraintAnd


class Port(IntegerType):
    subtypeSpec = ValueRange(10000, 65535)


class SequenceNumber(IntegerType):
    subtypeSpec = ValueRange(0, 65535)


class Message(OctetStringType):
    subtypeSpec = ValueSize(1, 4)


class UTC_Timestamp(SequenceType):
    class seconds(IntegerType):
        subtypeSpec = ValueRange(0, 4294967295)

    class useconds(IntegerType):
        subtypeSpec = ValueRange(0, 4294967295)

    rootComponent = AdditiveNamedTypes(
        NamedType('seconds', seconds()),
        NamedType('useconds', useconds()),
    )
    componentType = rootComponent


class Start(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('sequenceNumber', SequenceNumber()),
        NamedType('timestamp', UTC_Timestamp()),
        NamedType('srcPort', Port()),
        NamedType('dstPort', Port()),
    )
    componentType = rootComponent


class Stop(SequenceType):
    rootComponent = AdditiveNamedTypes(
        NamedType('sequenceNumber', SequenceNumber()),
        NamedType('timestamp', UTC_Timestamp()),
        NamedType('srcPort', Port()),
        NamedType('dstPort', Port()),
    )
    componentType = rootComponent


class Alive(SequenceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('timestamp', UTC_Timestamp()),
    )
    componentType = rootComponent


class Payload(SequenceOfType):
    subtypeSpec = SequenceOfValueSize(1, 5)
    componentType = Message()


class Data(SequenceType):
    class swRelease(EnumeratedType):
        subtypeSpec = ExtensionMarker(True)
        enumerationRoot = NamedValues(
            ('rel1', 0),
            ('rel2', 1),
            ('rel3', 2),
        )
        namedValues = enumerationRoot

    class macroId(BitStringType):
        subtypeSpec = ValueSize(20, 20)

    rootComponent = AdditiveNamedTypes(
        NamedType('sequenceNumber', SequenceNumber()),
        NamedType('swRelease', swRelease()),
        OptionalNamedType('macroId', macroId()),
        NamedType('payload', Payload()),
    )
    componentType = rootComponent


class SimpleMessage(ChoiceType):
    subtypeSpec = ExtensionMarker(True)
    rootComponent = AdditiveNamedTypes(
        NamedType('start', Start()),
        NamedType('stop', Stop()),
        NamedType('alive', Alive()),
        NamedType('data', Data()),
    )
    componentType = rootComponent


